
%if 0

RxANSI: Combined ANSI and LESS driver. AMIS v3.6 compliant.
Public Domain by Henrik Haftmann and Christian Masloch

RxANSI.ASM - Resident code of RxANSI

Read RxANSI.DOC for more information.

%endif

%if 0

Supported Int2F functions:
(not supported while ANSI functions switched off)

PTS-DOS - Get colour
INP:	ax = 1262h
OUT:	al = current video attribute (colour)
Gets the current colour, which is set by 2F.1263 or with ANSI functions.

PTS-DOS - Set colour
INP:	ax = 1263h
	dl = new video attribute (colour)
OUT:	-

ANSI.SYS - Installation check
INP:	ax = 1A00h
OUT:	al = 0FFh


Supported Int2D functions:

AMIS - Installation check
INP:	al = 00h
OUT:	al = 0FFh
	cx = Private version number (currently 010Ah)
	dx:di-> signature: "cm      ","RxANSI  "

AMIS - Get private entry point - NOP: no private entry point
INP:	al = 01h
OUT:	al = 00h

AMIS - Uninstall - NOP: no resident uninstaller or NOP: can't uninstall
INP:	al = 02h
OUT:	If installed from command line,
	 al = 03h
	 bx = memory block of resident RxANSI (cs)
	(If internal (pre-installed),
	 al = 01h
	If with resident uninstaller,
	 al = 01h if unsuccessful
	 al = 0FFh if successful)

AMIS - Request pop-up - NOP: no pop-up
INP:	al = 03h
OUT:	al = 00h

AMIS - Determine chained interrupts
INP:	al = 04h
OUT:	al = 04h
	dx:bx-> interrupt hook list (Int29, Int2D always. Optionally Int2F.)

AMIS - Get hotkeys - NOP: no hotkeys
INP:	al = 05h
OUT:	al = 00h

AMIS - Get device driver information - NOP: no device
INP:	al = 06h
OUT:	al = 00h

AMIS - Reserved for AMIS
INP:	al = 07h..0Fh
OUT:	al = 00h

RxANSI - Get compatible version
INP:	al = 10h
OUT:	al = 0FFh
	cx = compatible version number (currently 010Ah)

RxANSI - Get ctrl0
INP:	al = 20h
OUT:	al = size of returned control data (not 0)
	dx:bx-> control data
	LESS bit switched off if non-LESS version

RxANSI - Reserved for RxANSI
INP:	al = 11h..1Fh, 21h..FFh
OUT:	al = 00h

%endif

%if RXANSI_VERP != ver(1,10)		; if defined one doesn't match current
 %error "RxANSI version doesn't match"
%endif

handler:
				; AMIS stuff
amissig:
.ven:	fill 8,32,db "cm"		; vendor
.prod:	fill 8,32,db "RxANSI"		; product
.desc:	asciz ;"ANSI escape sequence parser"	; save memory by omitting the description

amisintr:
.i29:	db 29h
	dw i29
%if _I2F
.i2F:	db 2Fh
	dw i2F
%endif
.i2D:	db 2Dh
	dw i2D

				; RxANSI data
ctrl0:	db 1100_0000b			; control byte
					; bit 7: process ANSI sequences
					;     6: process LESS sequences
					;     5: last character was a digit
attrib:	db 7				; character attribute (colour)
last:	dw 0				; last two characters
argn:	db 0				; number of arguments
argv:	times 5 db 0			; argument values
curv:	dw 0				; saves cursor position

%if _I2F
				; i2F to prevent other ANSI(.SYS) from installing,
				; and to provide PTS-DOS get/set colour functions
iispentry i2F, 0, i2D
	test byte [cs:ctrl0], 80h	; ANSI active ?
	jz .chain			; no -->
	cmp ah, 12h			; 12h = DOS services ?
	je .colour			; yes, check for colour functions -->
	cmp ah, 1Ah			; 1Ah = ANSI.SYS services ?
	jne .chain			; no, chain -->
	test al, al			; zero = installation check
	jz .installationcheck
	cmp al, 02h			; both internal ANSI.SYS functions
	jbe .returnerror		;  we'll just return an unknown error -->
					; Like RBIL says about MS-DOS 5+ ANSI.SYS, chain on al>2
					; (Some MS-DOS ANSI.SYS are buggy and compare signed
					;  so they'll return with al=0 or so on al>=80h)
.chain:
	jmp far [cs:.next]

.colour:
	cmp al, 62h			; 1262h = Get colour ?
	je .getcolour			; yes --> (NC)
	cmp al, 63h			; 1263h = Set colour ?
	jne .chain			; no, chain --> (else NC)

.setcolour:
	mov byte [ cs:attrib ], dl	; set the colour
	jmp short .returnCF

.getcolour:
	mov al, byte [ cs:attrib ]	; get the colour
	jmp short .returnCF

.installationcheck:
	dec al				; = FFh
	iret

.returnerror:				; Set interrupt caller's flags to CY, return with ax=0
	xor ax, ax
	stc
.returnCF:
	push bp
	mov bp, sp
			; [bp +0]	= bp
			; [bp +2]	= ip
			; [bp +4]	= cs
			; [bp +6]	= fl
	pushf
	and byte [ bp+6 ], ~1		; NC
	popf
	jnc .NC
	inc byte [ bp+6 ]		; CY
.NC:
	pop bp
	iret
%endif

i2D.uninstall:
	inc ax				; (= 03h) safe to remove but no resident uninstaller
	mov bx, cs			; = segment
i2D.hwreset equ $-1		; (second byte of mov bx, cs is same as the retf opcode)
	iret

iispentry i2D, 0, i2D
	cmp ah, 0
amisnum equ $-1				; AMIS multiplex number (data for cmp opcode)
	je .handle			; our multiplex number -->
	jmp far [cs:.next]		; else go to next handler -->

.handle:
	test al, al
	jz .installationcheck		; installation check -->
	cmp al, 02h
	je .uninstall			; uninstallation -->
	cmp al, 04h
	je .determineinterrupts		; determine hooked interrupts -->
	cmp al, 10h
	je .getver			; get compatible version -->
	cmp al, 20h
	je .getctrl0			; get ctrl0 -->
				; all other functions are reserved or not supported by RxANSI
.nop:
	mov al, 0			; show not implemented
	iret

.installationcheck:
	dec al				; (= FFh) show we're here
	mov cx, RXANSI_VERP		; = version
	xor di, di			; dx:di -> AMIS signature strings of this program
.iret_dx_cs:
	mov dx, cs
.iret:
	iret

.determineinterrupts:			; al = 04h, always returns list
	mov bx, amisintr		; dx:bx -> hooked interrupts list
	jmp short .iret_dx_cs

.getver:
	mov al, 0FFh			; show call supported
	mov cx, RXANSI_VERC		; = compatible version
	iret

.getctrl0:
	mov bx, ctrl0			; dx:bx -> ctrl0
%ifn _LESS
	and byte [cs:bx], ~40h		; switch off LESS bit
%endif
	mov al, 1			; size of ctrl0: 1 byte
	jmp short .iret_dx_cs


iispentry i29, 80h, i2D
	sti
	push ax
	push bx
	push cx
	push dx
	push ds
	push es
	 push cs
	 pop ds

	mov dx, 40h
	mov es, dx			; BIOS data
	mov dx, word [last]
	mov bl, 7			; Default colour
	test byte [ctrl0], 80h		; ANSI active?
	jz .noansi			;  no -->

				; ANSI parser
	 mov bl, [attrib]
	 cmp dx, 27<<8|'['		; Escape,'[' ? (parsing an ANSI sequence ?)
	 jne .noansiseq			; no -->
	 call ansiseq
	 jmp short .exit		; ANSI sequence doesn't need to display anything
.noansiseq:
	 cmp al, 27			; Escape?
	 je .filtout
	 cmp dl, 27			; Former character was escape?
	 jne .ansichr			;  No, display normal character -->
	 cmp al, '['
	 je .filtout			; will start parsing the sequence on the next character
	 push ax
	  mov al, dl
	  call putcolour
	  mov ah, 0Eh			; TTY output, display the escape that didn't mark a sequence
	  int 10h
	 pop ax

.ansichr:
%if _LESS
	test byte [ctrl0], 40h		; Parse LESS additionally?
	jz .noless			;  No, display with ANSI colour -->

.noansi:
	test byte [ctrl0], 40h		; LESS active?
	jz .putout			;  No, display with default TTY output -->

				; LESS parser
.both:	cmp al, 32			; space ?
	je .noless			; doesn't make sense to change foreground colour -->
				; This fixes the old bug that the command line would display
				; a green/yellow cursor when a space/underscore was typed and
				; then the backspace key was used to erase that space/underscore.
				; (Internally, the command line handler overwrites the erased
				; character with a space.)
	cmp dl, 8			; Backspace?
	jne .noless			;  No overstruck sequence -->
	cmp dh, '_'			; Underscore - Backstep - Character? (underlined)
	jz .underline			;  Display yellow -->
	cmp dh, al			; Character - Backstep - Same character? (bold)
	jnz .noless			;  If not, default (ANSI) colour -->

.bold:				; "Bold" is displayed green
	and bl, 0F8h			; Zero foreground colour
	or bl, 2			; Add foreground colour green
colourbold equ ($-1)			; Colour for bold (data for or opcode)
	jmp short .noless

.underline:			; "Underline" is displayed yellow
	and bl, 0F8h			; Zero foreground colour
	or bl, 14			; Add foreground colour yellow
colourunderline equ ($-1)		; Colour for underline (data for or opcode)

.noless:			; No LESS sequence or fall-through
%else	; _LESS
.noansi equ .putout
%endif	; _LESS
	call putcolour
	cmp al, 10
	jnz .putout
	push dx
	 call getcursor			;PA: DX
	 cmp dh, [es:84h]		; last row?
	 jnz .putout_1			; no, continue normal
	  xor cx, cx
	  mov dl, [es:4Ah]		; last column
	  dec dl
	  mov bh, [attrib]
	  mov ax, 0601h
	  int 10h			; Scroll one line
	pop dx
	mov al, 10
	jmp short .filtout

.putout_1:
	pop	dx

.putout:			; Just display with default TTY output if
				;  both ANSI and LESS are inactive
	mov ah, 0Eh
	int 10h				; TTY output

.filtout:			; Don't display but still store into buffer
	 xchg dh, dl
	 mov dl, al
	 mov word [last], dx
.exit:
	pop es
	pop ds
	pop dx
	pop cx
	pop bx
	pop ax
	iret

putcolour:		; Display space with colour bl, al = actual character
	cmp al, 7
	jz .return
	cmp al, 8
	jz .return
	cmp al, 13
	jz .return
	cmp al, 10
	jz .return			; these don't need a colour
	push cx
	push ax
	 mov bh,[es:62h]		; Active video page
	 mov cx,1
	 mov ax,0920h			; write character and attribute at cursor position,
	 				;  character = blank
	 int 10h
	pop ax
	pop cx
.return:retn

ansiseq:		; Process ANSI sequence
	push si
%if _386
	movzx bx, byte [ argn ]
%else
	 mov bl, byte [ argn ]
	 xor bh, bh			; Fill bx with byte [argn]
%endif
	 add bx, argv - 1
	 sub al, '0'			; like cmp al,'0'
	 jb .noarg			;  if below -->
	 cmp al, 10			; Digit?
	 jnb .noarg			;  No -->
	 test byte [ ctrl0 ], 20h	; Last character was digit?
	 jnz .second_digit		;  Yes, create number from both digits -->
	 inc byte [ argn ]		;  No, create new argument with one digit
	 inc bx				; point to storage for new argument
	 cmp bx, byte argv+5		; too many arguments ?
	 jae .ret			; yes, ignore -->
	 jmp short .first_digit		;  And store digit -->

.second_digit:
	 mov ah, byte [ bx ]		; Get old digit
	 add ah, ah
	 add ah, ah
	 add ah, byte [ bx ]
	 add ah, ah			; Multiply old digit by ten: ((old*2*2)+old)*2
	 add al, ah			; add old digit to new digit's value
.first_digit:
	 mov byte [ bx ], al		; store number
	 or byte [ ctrl0 ], 20h
	 jmp short .ret
.noarg:
	 add al,'0'			; Restore previous value
	 call subfunc			; Parse letter if any
.ret:	pop si
	retn

		; OUT:	dx = cursor position of active video page
getcursor:
	push bx
%if _386
	movzx bx, byte [es:62h]	; Active video page
%else
	 mov bl,[es:62h]		; Active video page
	 xor bh, bh
%endif
	 add bx, bx
	 mov dx,[es:50h +bx]		; Cursor position inside BIOS data
	pop bx
	retn

subfunc:
	mov si, functiontable
	call dispatcher
	jc .end				; Unknown function
	call getcursor
	cmp byte [argn], 1		; CY if argn zero
	call word [si]
.end:	xor ax, ax
	mov word [ last ], ax		; Remove saved ANSI arguments
	mov word [ argn ], ax		; Reset argument counter
	mov word [ argv+1 ], ax		; Insure arguments are zero
	mov word [ argv+3 ], ax		; Insure arguments are zero
	and byte [ ctrl0 ], ~20h	; Mark end of string
	retn

functiontable:
		db ';'			; Ignore semicolon
		dw afsemi
		db '='			; Ignore equal sign
		dw afequal
		db 'H'			; Set cursor position
		dw afh
		db 'f'			; Same as 'H'
		dw afh
		db 'A'			; Move cursor up
		dw afa
		db 'B'			; Move cursor down
		dw afb
		db 'C'			; Move cursor forward (right)
		dw afc
		db 'D'			; Move cursor backward (left)
		dw afd
		db 's'			; Save cursor position
		dw afs
		db 'u'			; Restore cursor position
		dw afu
		db 'J'			; Clear screen, reset cursor position
		dw afj
		db 'K'			; Clear line from cursor position
		dw afk
		db 'm'			; Set video colours
		dw afm
		db 'h'			; Set video mode
		dw afx
		db 'l'			; Reset video mode
		dw afx
		db -1

		; Helper function that adjusts stack
afsemi:
	and byte [ctrl0], ~20h		; end of string
afequal:
	pop si				; adjust stack
	retn				; level above

		; Cursor functions
afa:
	sbb	dh, [argv]
	jnc	setcur
	xor	dh, dh
	jmp short setcur

afb:
	adc	dh, [argv]
	jc	.overflow
	cmp	dh, [es:84h]		; higher than last row ?
	jc	setcur			; no -->
.overflow:
	mov	dh, [es:84h]		; overflow, force to last row
	jmp short setcur

afc:
	adc	dl, [argv]
	jc	.overflow
	cmp	dl, [es:4Ah]		; higher than last column ?
	jc	setcur			; no -->
.overflow:
	mov	dl, [es:4Ah]
	dec	dl			; overflow, force to last column
	jmp short setcur

afd:
	sbb	dl, [argv]
	jnc	setcur
	xor	dl, dl			; underflow, force to zero
	jmp short setcur

afh:
	mov	dx, word [argv]
	xchg	dh, dl			; Need to exchange
	cmp	dh, [es:84h]
	ja	.ret
	cmp	dl, [es:4Ah]
	jb	setcur			; valid, set cursor position -->
.ret:
	retn

		; Clear row or screen
afj:
				; Here was a check if byte [argv] is 2. Removed because
				; ANSI.SYS in fact doesn't check, and it's shorter.
	xor	cx, cx
	mov	dl, [es:4Ah]		; number of rows
	dec	dl
	mov	dh, [es:84h]		; number of columns
	mov	bh, [attrib]
	mov	ax, 0600h
	int	10h			; Clear screen
	xor	dx, dx			; Reset cursor to 0/0

setcur:
	mov	bh, [es:62h]
	mov ah, 02h
	int 10h				; Set cursor
	retn

afk:
	mov	cx, dx
	mov	dl, [es:4Ah]		; number of columns
	dec	dl
	mov	bh, [attrib]
	mov ax, 0600h
	int 10h				; Clear rest of line
	retn


		; Store and restore cursor position
afs:
	mov word [curv], dx		; store cursor position
	retn

afu:
	mov dx, word [curv]		; restore
	jmp short setcur


		; Select colour and attributes
afm:
	mov	bx, argv		; preset to argv
	mov	cl, [argn]
	xor	ch, ch
	jcxz	.ret
.loop:
	 mov	al, [bx]
	 cmp	al, 30
	 jc	.nofgcolour
	 cmp	al, 38
	 jnc	.nofgcolour
	 sub	al, 30
	 push	bx
	  mov	bx, BRevTab
	  xlat
	 pop	bx
	 mov	ah, ~7
	 jmp short .mani		; New foreground colour
.nofgcolour:
	 cmp	al, 40
	 jc	.nobgcolour
	 cmp	al, 48
	 jnc	.nobgcolour
	 sub	al, 40
	 push	bx
	  mov	bx, BRevTab
	  xlat
	 pop	bx
%if _186
	 rol	al, 4
%else
	 push	cx
	  mov	cl, 4
	  rol	al, cl			; Exchange high and low nibble
	 pop	cx
%endif
	 mov	ah, ~70h
.mani:
	 call	atMani
	 jmp short .exit
.nobgcolour:
	 mov	si, subfunctiontable
	 call	dispatcher
	 jc	.exit
	 call	word [si]
.exit:
	inc	bx			; next argument
	loop	.loop
.ret:
	retn

BRevTab: 		; table to get colour number
	db	0,4,2,6,1,5,3,7

		; sub functions
atReset:
	mov byte [attrib], 7			; reset to default colour
	retn

atIntens:
	or byte [attrib], 8		; set intensity
	retn

atUnderl:
	mov ax, 0F801h			; blue for underline
	jmp short atMani		; change attribute

atFlash:
	or byte [attrib], 80h
	retn

atInvert:
	mov al, byte [attrib]
	and al, 77h			; separate both colours
	mov ah, 88h			; change both colours
atRotMani:
	rol al, 1
	rol al, 1
	rol al, 1
	rol al, 1			; exchange colours, or shift background colour down
atMani:
	and ah, byte [attrib]		; mask off old colour
	or al, ah
	mov byte [attrib], al		; new colour
	retn

atHide:
	mov al, byte [attrib]
	and al, 70h			; separate background colour
	mov ah, 0F0h			; change foreground colour only, clear intensity
	jmp short atRotMani

subfunctiontable:
	db 0
	dw atReset
	db 1
	dw atIntens
	db 4
	dw atUnderl
	db 5
	dw atFlash
	db 7
	dw atInvert
	db 8
	dw atHide
	db -1

		; Select video mode
afx:
	jc .ret				; if argn=0
%if _386
	movzx ax, byte [argv]
%else
	mov al, [argv]
	mov ah, 0
%endif
	int 10h
.ret:	retn


dispatcher.loop:
	add si, byte 3			; next table entry
dispatcher:			; dispatcher
	cmp byte [si], al
	jz .found
	cmp byte [si], 0FFh		; End of table? (CY if below)
	cmc				; CY if at end of table (not below)
	jnc .loop			; else NC (value is below) -->
.found:	inc si				; wo[si] == function pointer
	retn

	align 16
	endarea handler
